#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/mman.h>
#include <fcntl.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>

#include <X11/extensions/Xv.h>
#include <X11/extensions/Xvlib.h>

#include <SDL/SDL.h>

#include "re.h"
#include "xvmc_tests.h"

void free_xvmc_state(struct xvmc_state_t* state)
{
	if (state->inited == 0) return;

	Display *dpy = state->sdli.info.x11.display;

	XvMCDestroyMacroBlocks(dpy, &state->mv_blocks);
	XvMCDestroyBlocks(dpy, &state->data_blocks);
	XvMCDestroySurface(dpy, &state->surface);
	XvMCDestroyContext(dpy, &state->ctx);
	if (state->port != 0) {
		XvUngrabPort(dpy, state->port, CurrentTime);
		state->port = 0;
	}
}

Atom xv_intern_atom(Display *dpy, unsigned long port, const char *name)
{
	XvAttribute *attrs;
	int num_attrs, i;
	Atom atom = None;

	attrs = XvQueryPortAttributes(dpy, port, &num_attrs);
	if (attrs == NULL) return atom;
	for (i = 0; i < num_attrs; i++) {
		if (!strcmp(attrs[i].name, name)) {
			atom = XInternAtom(dpy, name, False);
			break;
		}
	}
	XFree(attrs);
	return atom;
}

/* Initialization code heavily borrowed from vo_xvmc.c in MPlayer */
int init_xvmc_state(struct xvmc_state_t* state, int width, int height)
{
	int i, j, k, res;
	unsigned int adaptors;
	XvAdaptorInfo *adaptor_info;
	int num_surface_infos;

	state->inited = 0;

	if (width > 16 || height > 16) {
		/* Only support a single block */
		return -1;
	}

	SDL_VERSION(&state->sdli.version);
	if (!SDL_GetWMInfo(&state->sdli)) {
		return -1;
	}

	state->sdli.info.x11.lock_func();

	Display *dpy = state->sdli.info.x11.display;

	state->width = width;
	state->height = height;
	res = XvQueryAdaptors(dpy, DefaultRootWindow(dpy),
			      &adaptors, &adaptor_info);
	if (res != Success) {
		return -1;
	}

	for (i = 0; i < adaptors; i++) {
		if (adaptor_info[i].type == 0) {
			continue;
		}

		for (j = 0; j < adaptor_info[i].num_ports; j++) {
			XvMCSurfaceInfo *surface_info;
			unsigned long port = adaptor_info[i].base_id + j;
			surface_info = XvMCListSurfaceTypes(dpy, port,
							    &num_surface_infos);
			if (surface_info == NULL || num_surface_infos == 0) {
				continue;
			}

			for (k = 0; k < num_surface_infos; k++) {
				if (width > surface_info[k].max_width ||
				    height > surface_info[k].max_height) {
					continue;
				}
				res = XvGrabPort(dpy, port, CurrentTime);
				if (res != Success) continue;
				state->port = port;
				memcpy(&state->surface_info, &surface_info[k],
				       sizeof(XvMCSurfaceInfo));
				printf("mc_type: %d\n", state->surface_info.mc_type);
				break;
			}

			XFree(surface_info);

			if (k != num_surface_infos) goto surface_filled;
		}
	}
 surface_filled:

	XFree(adaptor_info);

	res = XvMCCreateContext(dpy, state->port,
				state->surface_info.surface_type_id,
				width, height, XVMC_DIRECT, &state->ctx);
	if (res != Success) {
		fprintf(stderr, "Could not create an xvmc context");
		return -1;
	}

	if (!(state->ctx.flags & XVMC_DIRECT)) {
		fprintf(stderr, "Did not allocate a direct xvmc context.\n");
		return -1;
	}

	state->num_blocks = 1; /* XXX */

	if (state->surface_info.chroma_format != XVMC_CHROMA_FORMAT_420) {
		fprintf(stderr, "Expected a 4:2:0 surface\n");
		return -1;
	}

	int blocks_per_macroblock = 6; /* XXX Assumes 4:2:0 */
	res = XvMCCreateBlocks(dpy, &state->ctx,
			       state->num_blocks * blocks_per_macroblock,
			       &state->data_blocks);
	if (res != Success) {
		XvMCDestroyContext(dpy, &state->ctx);
		return -1;
	}

	res = XvMCCreateMacroBlocks(dpy, &state->ctx, state->num_blocks,
				    &state->mv_blocks);
	if (res != Success) {
		XvMCDestroyBlocks(dpy, &state->data_blocks);
		XvMCDestroyContext(dpy, &state->ctx);
		return -1;
	}

	res = XvMCCreateSurface(dpy, &state->ctx, &state->surface);
	if (res != Success) {
		XvMCDestroyMacroBlocks(dpy, &state->mv_blocks);
		XvMCDestroyBlocks(dpy, &state->data_blocks);
		XvMCDestroyContext(dpy, &state->ctx);
		return -1;		
	}

	Atom xv_atom;
	// This ended up as "colorkey not found" so code disabled.
	/*
	xv_atom = xv_intern_atom(state->dpy, state->port, "XV_COLORKEY");
	if (xv_atom == None) {
		fprintf(stderr, "Colorkey not found.\n");
		return -1;
	}

	res = XvGetPortAttribute(state->dpy, state->port, xv_atom,
				 &state->colorkey);
	if (res != Success) {
		return -1;
	}

	xv_atom = xv_intern_atom(state->dpy, state->port,
				 "XV_AUTOPAINT_COLORKEY");
	if (xv_atom == None) {
		fprintf(stderr, "Couldn't get autopaint colorkey setting.\n");
		return -1;
	}

	res = XvSetPortAttribute(state->dpy, state->port, xv_atom, 1);
	if (res != Success) {
		fprintf(stderr, "Couldn't turn colorkey autopaint on.\n");
		return -1;
	}
	*/

	xv_atom = xv_intern_atom(dpy, state->port, "XV_SYNC_TO_VBLANK");
	if (xv_atom != None) {
		/* Don't care about failure */
		XvSetPortAttribute(dpy, state->port, xv_atom, 1);
	}

	state->sdli.info.x11.unlock_func();

	state->inited = 1;

	return 0;
}

int init_xvmc_window(struct xvmc_state_t *state)
{
	state->sdli.info.x11.lock_func();

	Display *dpy = state->sdli.info.x11.display;
	Window win = state->sdli.info.x11.window;

	Window root = DefaultRootWindow(dpy);
	XVisualInfo vinfo;
	XSetWindowAttributes xswa;
	XWindowAttributes attribs;
	unsigned long xswa_mask;

	XGetWindowAttributes(dpy, root, &attribs);
	printf("Root window depth: %d\n", attribs.depth);

	/* Should be colorkey but my FX5200 doesn't seem to use that. */
	xswa.background_pixel = 0;
	xswa.border_pixel = 0;
	xswa_mask = CWBackPixel | CWBorderPixel;

	if (XMatchVisualInfo(dpy, DefaultScreen(dpy),
			     attribs.depth, TrueColor,
			     &vinfo) == 0) {
		fprintf(stderr, "Unable to get true color visual\n");
		return -1;
	}

	XUnmapWindow(dpy, win);
	XChangeWindowAttributes(dpy, win, xswa_mask, &xswa);
	XMapWindow(dpy, win);

	state->sdli.info.x11.unlock_func();

	return 0;
}

void print_pages(char *old, char *new, int length)
{
	int i;
	printf("Before:");
	for (i = 0; i < length; i++) {
		if (i % 26 == 0) printf("\n");
		printf("%02hhX ", old[i]);
	}

	printf("\n" "After:");
	for (i = 0; i < length; i++) {
		if (i % 26 == 0) printf("\n");
		printf("%02hhX ", new[i]);
	}
	printf("\n");
}

void cmp_mem(void *old, void *new, size_t length)
{
	int page = 4096; /* FIXME: hardcoded page size, use sysconf() */
	size_t i, j;
	for (i = 0; i < length / page; i++) {
		int *tmp_old = old + i * page;
		int *tmp_new = new + i * page;
		for (j = 0; j < page / sizeof(int); j++) {
			if (tmp_old[j] != tmp_new[j]) {
				//printf("==================================\n");
				printf("AGP [0x%lX]: 0x%08X -> 0x%08X\n",
				       i * page + j * sizeof(int),
				       tmp_old[j], tmp_new[j]);
				//print_pages(tmp_old, tmp_new, page);
				break;
			}
		}
	}
}

void test_idct_intra(struct xvmc_state_t *state)
{
	int i;

	Display *dpy = state->sdli.info.x11.display;
	Window win = state->sdli.info.x11.window;

	state->sdli.info.x11.lock_func();

	redirect_output("test_idct_intra");

	/* Fill the 6 data_blocks with decreasing data, for easy recogntion */
	printf("4:2:0 macroblocks for a 16x16 image:");
	for (i = 0; i < state->data_blocks.num_blocks * 64; i++) {
		state->data_blocks.blocks[i] = 0x7FF0 - (i << 6);
		if (i % 8 == 0)
			printf("\n");
		if (i % 64 == 0)
			printf("New block:\n");
		printf("0x%4hx ", state->data_blocks.blocks[i]);
	}
	printf("\n");

	/* Set the macro block up to decode a frame (not field) intra block */
	XvMCMacroBlock *block = state->mv_blocks.macro_blocks;
	block->x = block->y = 0;
	block->macroblock_type = XVMC_MB_TYPE_INTRA;
	block->motion_type = 0;
	block->motion_vertical_field_select = 0;
	memset(block->PMV, 0, sizeof(block->PMV));
	block->index = 0;
	block->coded_block_pattern = 0xFFFF;
	block->dct_type = XVMC_DCT_TYPE_FRAME;

	printf("rendering surface\n");
	//sleep(5);
	/*int fd = open("/dev/mem", O_RDONLY);
	int length = 32 * 1024 * 1024;
	size_t offset = 96 * 1024 * 1024;
	void *agp_mem = mmap(NULL, length, PROT_READ, MAP_SHARED,
			     fd, 0xd0000000UL + offset);
	void *agp_copy1 = malloc(length);
	void *agp_copy2 = malloc(length);
	memcpy(agp_copy1, agp_mem, length);*/
	//dump_before();
	//dumpreg_before();
	if (XvMCRenderSurface(dpy, &state->ctx, XVMC_FRAME_PICTURE,
			      &state->surface, NULL, NULL,
			      0, 1, 0, &state->mv_blocks,
			      &state->data_blocks) != Success) {
		printf("Couldn't render surface.\n");
		return;
	}
	//dumpreg_after();
	//dump_after(0);

	/*memcpy(agp_copy2, agp_mem, length);
	cmp_mem(agp_copy1, agp_copy2, length);
	munmap(agp_mem, length);
	close(fd);
	free(agp_copy1);
	free(agp_copy2);*/

	printf("syncing surface\n");
	//dump_before();
	//dumpreg_before();
	XvMCSyncSurface(dpy, &state->surface);
	//dumpreg_after();
	//dump_after(0);

	printf("displaying surface\n");
	//dump_before();
	//dumpreg_before();
	if (XvMCPutSurface(dpy, &state->surface, win,
			   0, 0, 16, 16,
			   0, 0, 512, 512,
			   XVMC_FRAME_PICTURE) != Success) {
		printf("Couldn't display rendered surface.\n");
		return;
	}
	//dumpreg_after();
	//dump_after(0);

	printf("flushing display\n");
	//dump_before();
	//dumpreg_before();
	XFlush(dpy);
	//dumpreg_after();
	//dump_after(0);
	close_output();

	state->sdli.info.x11.unlock_func();

	sleep(1);
}
